import base64
import functools
import glob
import importlib_metadata
import inspect
import os
import shutil
import signal
import sys
import threading
import types
import uuid

import cloudpickle
import docker
import tempfile
import toolchest_client

from .containers import DockerContainer
from .module_detection import get_modules_to_register


def patch_system_call(user_docker_container_name=None, original_function=None, pass_kwargs=True,
                      docker_shell_location=None):
    """
    Patch os.system, subprocess.run, or subprocess.Popen
    Note: docker_shell_location refers to the shell in the *user* docker container
    """

    def run(*args, **kwargs):
        """Lug-replaced function."""
        docker_exec_args = [
            "docker",
            "exec",
            f"{user_docker_container_name}",
        ]
        using_shell = kwargs.get("shell") or original_function.__name__ == "system"
        if using_shell:
            docker_exec_args.append(docker_shell_location)

        if type(args[0]) is list:
            original_list_args = args[0]
            docker_exec_args += original_list_args
            return original_function(docker_exec_args, *args[1:], **kwargs)
        else:
            # NOTE: args[0] (the user command argument) can only be a string if using_shell is True.
            stringified_docker_command = ' '.join(docker_exec_args)
            user_command = args[0].replace("'", r"'\''")  # sanitizes single quotes within sh command
            new_args = stringified_docker_command + " -c '" + user_command + "'"
            if pass_kwargs:
                return original_function(new_args, **kwargs)
            return original_function(new_args)

    run.is_lug_function = True
    run.lug_original_function = original_function
    return run


def find_and_replace_function(member, unique_docstring, replace_with, patched_functions, depth=0, max_depth=3):
    if depth > max_depth:
        return
    if type(member) in [str, type(None)]:
        return
    elif type(member) == dict:
        for name, sub_attribute in member.copy().items():  # The dict can change size, so make a copy for iteration
            if name == "__builtins__":
                continue
            this_one = find_and_replace_function(
                member=sub_attribute,
                unique_docstring=unique_docstring,
                replace_with=replace_with,
                patched_functions=patched_functions,
                depth=depth + 1,
                max_depth=max_depth,
            )
            if this_one:
                patched_functions.append({
                    "member": member,
                    "attribute_name": name,
                    "original_value": replace_with.lug_original_function,
                    "is_dictionary": True,
                })
                member[name] = replace_with
                return False
    elif isinstance(member, types.ModuleType):
        # import subprocess -> subprocess is a module
        # import os -> os is a module
        try:
            members = inspect.getmembers(member)
        # For some members, inspect.getmembers returns "TypeError: '_ClassNamespace' object is not iterable"
        except (ModuleNotFoundError, TypeError):
            return
        for name, sub_member in members:
            if name == "__builtins__":
                continue
            this_one = find_and_replace_function(
                member=sub_member,
                unique_docstring=unique_docstring,
                replace_with=replace_with,
                patched_functions=patched_functions,
                depth=depth + 1,
                max_depth=max_depth,
            )
            if this_one:
                patched_functions.append({
                    "member": member,
                    "attribute_name": name,
                    "original_value": getattr(member, name),
                    "is_dictionary": False,
                })
                setattr(member, name, replace_with)
                return False
        return False
    else:
        try:
            if hasattr(member, "__doc__") and member.__doc__ and unique_docstring in member.__doc__:
                return True
            elif isinstance(member, types.FunctionType):
                # Explore the globals of any imported function
                find_and_replace_function(
                    member=member.__globals__,
                    unique_docstring=unique_docstring,
                    replace_with=replace_with,
                    patched_functions=patched_functions,
                    depth=depth + 1,
                    max_depth=5,  # If we're exploring a function globals, increase max depth to five
                )
                return False
        except (TypeError, ModuleNotFoundError):
            return False
        return False


def patch_system_calls(func, user_docker_container_name, docker_shell_location):
    # Create a new subprocess.run
    import subprocess
    new_subprocess_run = patch_system_call(
        user_docker_container_name=user_docker_container_name,
        original_function=subprocess.run,
        docker_shell_location=docker_shell_location,
    )
    # Create a new subprocess.Popen
    new_subprocess_popen = patch_system_call(
        user_docker_container_name=user_docker_container_name,
        original_function=subprocess.Popen,
        docker_shell_location=docker_shell_location,
    )
    # Create a new os.system
    import os
    new_os_system = patch_system_call(
        user_docker_container_name=user_docker_container_name,
        original_function=os.system,
        pass_kwargs=False,
        docker_shell_location=docker_shell_location,
    )

    patched_functions = []
    find_and_replace_function(
        member=func.__globals__,
        unique_docstring="Execute a child program in a new process.",
        replace_with=new_subprocess_popen,
        patched_functions=patched_functions,
    )
    if not hasattr(subprocess.Popen, "is_lug_function"):
        find_and_replace_function(
            member=func.__globals__,
            unique_docstring="Run command with arguments and return a CompletedProcess instance",
            replace_with=new_subprocess_run,
            patched_functions=patched_functions,
        )
    find_and_replace_function(
        member=func.__globals__,
        unique_docstring="Execute the command in a subshell",
        replace_with=new_os_system,
        patched_functions=patched_functions,
    )

    return patched_functions


def unpatch_system_calls(patched_functions):
    for patched_function_info in patched_functions:
        if patched_function_info["is_dictionary"]:
            attribute = patched_function_info["attribute_name"]
            patched_function_info["member"][attribute] = patched_function_info["original_value"]
        else:
            setattr(
                patched_function_info["member"],
                patched_function_info["attribute_name"],
                patched_function_info["original_value"],
            )


def can_be_pickled(module):
    # The module must be referenced inside the function to actually test if it can be pickled
    dummy_function_using_module = lambda x: module.__name__
    try:
        cloudpickle.register_pickle_by_value(module)
        cloudpickle.dumps(dummy_function_using_module)
        return True
    except Exception:
        return False
    finally:
        cloudpickle.unregister_pickle_by_value(module)


def find_module_transferability(modules):
    copyable_packages = set()
    uncopyable_packages = set()
    uncopyable_pip_names = dict()  # package_name: version # todo: filter out packages that can't be installed with pip
    packages_distributions = importlib_metadata.packages_distributions()
    # Split pickleable and unpickleable packages
    for module in modules:
        if not module.__spec__:
            continue

        picklable = can_be_pickled(module)

        # Find physical location of module source
        parent = module.__spec__.name.rpartition(".")[0]
        child = None
        if parent:
            child = module
            module = sys.modules[parent]
        submodule_search_locations = module.__spec__.submodule_search_locations
        loader_members = inspect.getmembers(module.__spec__.loader)
        loader_keys = [loader[0] for loader in loader_members]
        loader_path = None
        if "path" in loader_keys:
            loader_path = module.__spec__.loader.path
        # todo: handle multiple submodule search locations
        primary_location = (submodule_search_locations and submodule_search_locations[0]) or loader_path

        has_so_files = False
        if primary_location:
            adjusted_primary_location = os.path.dirname(primary_location) if primary_location.endswith(".py") \
                else primary_location

            # Check if the package contains any CPython compiled files (e.g. numpy, scipy, etc)
            has_so_files = glob.glob(f"{adjusted_primary_location}/*.so") \
                + glob.glob(f"{adjusted_primary_location}/**/*.so", recursive=True)
            if has_so_files:
                picklable = False

        if not picklable:
            try:
                pip_names = packages_distributions[module.__name__]
                for pip_name in pip_names:
                    # Drop local version identifiers (e.g. "+cu117" in "torch==1.13.1+cu117")
                    public_version = module.__version__.split("+")[0]
                    uncopyable_pip_names[pip_name] = public_version
                uncopyable_packages.add(module.__name__)
                # Make sure we know not to copy the child if present (e.g. lug and lug.run)
                if child:
                    uncopyable_packages.add(child.__name__)
            except KeyError:
                if not has_so_files:
                    # Add to packages that need to be manually copied
                    copyable_packages.add(primary_location)
                else:
                    raise ModuleNotFoundError(f"'{module.__name__}' distribution info not found on this machine.")
            continue

    return uncopyable_packages, uncopyable_pip_names, copyable_packages


def env_requirements_to_string(pip_packages):
    requirements_string = ""
    for name, version in pip_packages.items():
        requirements_string += f"{name}=={version} "
    return requirements_string


def create_python_script(func, args, kwargs, temp_input, user_docker, docker_shell_location, serialize_dependencies,
                         internal_dir):
    output_uuid = uuid.uuid4()
    with open(temp_input.name, 'w') as fp:
        copyable_packages = []
        pip_packages_string = ''
        links = []
        if serialize_dependencies:
            modules_to_register = get_modules_to_register(func)
            uncopyable_packages, uncopyable_pip_names, copyable_packages = find_module_transferability(
                modules_to_register
            )
            pip_packages_string = env_requirements_to_string(uncopyable_pip_names)
            pickle_packages = list(filter(
                lambda mod: mod.__name__ not in copyable_packages and mod.__name__ not in uncopyable_packages,
                modules_to_register,
            ))
            internal_dir_basename = os.path.basename(internal_dir)
            for module_location_to_include in copyable_packages:
                #  todo: handle conflicts for package directory names
                local_path_with_internal_dir = os.path.join(
                    internal_dir_basename,
                    os.path.basename(module_location_to_include)
                )
                abs_path_with_internal_dir = os.path.join(internal_dir, os.path.basename(module_location_to_include))
                if not os.path.exists(abs_path_with_internal_dir):
                    shutil.copytree(module_location_to_include, abs_path_with_internal_dir)
                links.append((module_location_to_include, f"./input/{local_path_with_internal_dir}"))
            for module in pickle_packages:
                cloudpickle.register_pickle_by_value(module)
        else:
            cloudpickle.register_pickle_by_value(sys.modules[__name__])
        pickled_func = cloudpickle.dumps(func)
        encoded_func = base64.encodebytes(pickled_func)
        pickled_patch = cloudpickle.dumps(patch_system_calls)
        encoded_pickled_patch = base64.encodebytes(pickled_patch)
        pickled_args = cloudpickle.dumps(args)
        encoded_args = base64.encodebytes(pickled_args)
        pickled_kwargs = cloudpickle.dumps(kwargs)
        encoded_kwargs = base64.encodebytes(pickled_kwargs)
        if serialize_dependencies:
            for module in pickle_packages:
                cloudpickle.unregister_pickle_by_value(module)
        else:
            cloudpickle.unregister_pickle_by_value(sys.modules[__name__])
        fp.write("import base64\n")
        fp.write("import cloudpickle\n")
        fp.write("import os\n")
        fp.write(f"for dir in {list(copyable_packages)}:\n")  # Make directories
        fp.write("\tos.makedirs(os.path.dirname(dir), exist_ok=True)\n")
        fp.write(f"for link in {links}:\n")  # ln all paths
        fp.write("\tos.symlink(os.path.join(os.getcwd(), link[1]), link[0])\n")
        fp.write(f"func = cloudpickle.loads(base64.decodebytes({encoded_func}))\n")
        fp.write(f"patch_system_calls = cloudpickle.loads(base64.decodebytes({encoded_pickled_patch}))\n")
        fp.write(f"args = cloudpickle.loads(base64.decodebytes({encoded_args}))\n")
        fp.write(f"kwargs = cloudpickle.loads(base64.decodebytes({encoded_kwargs}))\n")
        fp.write(f"patch_system_calls(func, '{user_docker.container_name}', '{docker_shell_location}')\n")
        fp.write("result = func(*args, **kwargs)\n")
        fp.write(f"with open(f'./output/encoded_output_{output_uuid}', 'wb') as file:\n")
        fp.write("\tfile.write(base64.encodebytes(cloudpickle.dumps(result)))\n")
    return output_uuid, pip_packages_string


def parse_toolchest_run(output_path, output_uuid):
    encoded_output_path = f'{output_path}/encoded_output_{output_uuid}'
    if os.path.exists(encoded_output_path):
        with open(encoded_output_path, 'rb') as file:
            encoded_result = file.read().replace(b'\n', b'')
            if encoded_result:
                result = cloudpickle.loads(base64.decodebytes(encoded_result))
            else:
                result = None
        os.remove(encoded_output_path)
    return result


def execute_remote(func, args, kwargs, toolchest_key, remote_output_directory, tmp_dir, image, remote_inputs,
                   user_docker, remote_instance_type, volume_size, python_version, docker_shell_location,
                   serialize_dependencies, command_line_args, streaming_enabled):
    if not os.path.exists(tmp_dir):
        os.mkdir(tmp_dir)
    # Create a .lug-uuid4()/ directory for Lug module propagation
    lug_internal_dir = os.path.join(tmp_dir, f".lug-{uuid.uuid4()}")
    os.mkdir(lug_internal_dir)

    temp_input = tempfile.NamedTemporaryFile(dir=tmp_dir)
    temp_directory = None
    output_uuid, pip_packages_string = create_python_script(
        func=func,
        args=args,
        kwargs=kwargs,
        temp_input=temp_input,
        user_docker=user_docker,
        docker_shell_location=docker_shell_location,
        serialize_dependencies=serialize_dependencies,
        internal_dir=lug_internal_dir,
    )
    try:
        if toolchest_key:
            toolchest_client.set_key(toolchest_key)
        if remote_output_directory is None:
            temp_directory = tempfile.TemporaryDirectory(dir=tmp_dir)
        output_directory = remote_output_directory or temp_directory.name
        if isinstance(command_line_args, list):
            command_line_args = " ".join(command_line_args)
        remote_inputs = remote_inputs or []
        remote_inputs.append(lug_internal_dir)
        remote_run = toolchest_client.lug(
            custom_docker_image_id=image,
            tool_version=python_version,
            inputs=remote_inputs,
            output_path=output_directory,
            script=temp_input.name,
            container_name=user_docker.container_name,
            docker_shell_location=docker_shell_location,
            tool_args=command_line_args,
            instance_type=remote_instance_type,
            volume_size=volume_size,
            streaming_enabled=streaming_enabled,
            compress_inputs=True,
            pip_dependencies=pip_packages_string,
            retain_base_directory=True,
        )
        status_response = remote_run.get_status(return_error=True)
        status = status_response['status']
        if status == 'failed':
            raise ValueError(f"Remote execution failed due to: {status_response['error_message']}")
        result = parse_toolchest_run(output_directory, output_uuid)
        return result
    finally:
        if temp_directory is not None:
            temp_directory.cleanup()
        if temp_input is not None:
            temp_input.close()


def execute_local(mount, client, user_docker, func, args, kwargs, docker_shell_location):
    # todo: make sure the docker containers don't exit shortly after spawn, propagate errors
    mount = os.path.realpath(mount)
    # Run user container
    user_docker.container = client.containers.run(
        image=user_docker.image_name_and_tag,
        volumes=[f'{mount}:/lug'],
        detach=True,
        stdin_open=True,
        name=user_docker.container_name,
        command=docker_shell_location,
        working_dir="/lug",
    )
    if threading.current_thread() is threading.main_thread() and not sys.platform.startswith('win'):
        # Only supports Unix signals
        signal.signal(signal.SIGABRT, user_docker.signal_kill_handler)
        signal.signal(signal.SIGHUP, user_docker.signal_kill_handler)
        signal.signal(signal.SIGSEGV, user_docker.signal_kill_handler)
        signal.signal(signal.SIGTERM, user_docker.signal_kill_handler)
    patched_functions = patch_system_calls(func, user_docker.container_name, docker_shell_location)
    try:
        result = func(*args, **kwargs)
    finally:
        unpatch_system_calls(patched_functions)
    return result


def run(image, mount=os.getcwd(), tmp_dir=tempfile.gettempdir(), docker_shell_location="/bin/sh", remote=False,
        remote_inputs=None, remote_output_directory=None, toolchest_key=None, remote_instance_type=None,
        volume_size=None, serialize_dependencies=False, command_line_args="", streaming_enabled=True):
    def decorator_lug(func):
        @functools.wraps(func)
        def inner(*args, **kwargs):
            # Find current Python version
            python_version = f"{sys.version_info.major}.{sys.version_info.minor}"
            supported_versions = ["3.8", "3.9", "3.10", "3.11"]
            if python_version not in supported_versions:
                raise ValueError(f"Python version {python_version} is not supported. PRs welcome!")
            user_docker = None
            try:
                if remote:
                    user_docker = DockerContainer(
                        docker_client=None,
                        image_name_and_tag=image,
                    )
                    result = execute_remote(
                        func=func, args=args, kwargs=kwargs, image=image, remote_inputs=remote_inputs,
                        toolchest_key=toolchest_key, remote_output_directory=remote_output_directory, tmp_dir=tmp_dir,
                        user_docker=user_docker, remote_instance_type=remote_instance_type, volume_size=volume_size,
                        python_version=python_version, docker_shell_location=docker_shell_location,
                        serialize_dependencies=serialize_dependencies, command_line_args=command_line_args,
                        streaming_enabled=streaming_enabled,
                    )
                else:
                    client = docker.from_env()

                    # Get or pull the user Docker image from local/remote
                    user_docker = DockerContainer(
                        docker_client=client,
                        image_name_and_tag=image,
                    )
                    user_docker.load_image(remote=remote)
                    result = execute_local(
                        func=func, args=args, kwargs=kwargs, mount=mount, client=client, user_docker=user_docker,
                        docker_shell_location=docker_shell_location
                    )
            finally:
                if user_docker is not None and user_docker.container is not None:
                    user_docker.container.reload()
                    if user_docker.container.status == "running":
                        user_docker.container.stop()
                    user_docker.container.remove()
            return result

        return inner

    return decorator_lug



import ast
import glob
import inspect
import sys
import types
import os

from .constants import STDLIB_MODULE_NAMES


def parse_ast_for_file(python_file_path):
    opened_file = open(python_file_path).read()
    parsed_ast = ast.parse(opened_file)
    return [node for node in ast.walk(parsed_ast)]


def get_registered_module_imports(modules_to_register, modules_to_skip):
    """
    Perform a deep search for module dependencies, by parsing the AST and looking for all imports.

    All imports in the module are parsed, even if they aren't used. Practically, this means we're catching
    more dependencies than needed within a module.

    Modules that this function finds are not searched for _their_ dependencies. For a truly deep search, this would need
    to be modified to search through found modules and their dependencies.

    Only modules that are already loaded within `sys.modules` of the Lugged function are registered for pickling.
    """
    # Find all sys level registered modules that aren't builtins
    filtered_sys_modules = []
    for module_name, module_value in sys.modules.items():
        is_builtin_module = module_name in sys.builtin_module_names
        is_stdlib_module = module_name in STDLIB_MODULE_NAMES
        if not (is_builtin_module or is_stdlib_module):
            filtered_sys_modules.append(module_name)

    # Find all imported modules within dependencies
    deep_modules = set()
    for module in modules_to_register:
        all_ast_nodes = []

        # Some modules are single-file modules, or otherwise don't have a root directory
        if not hasattr(module, "__path__"):
            # We're not supporting modules that have neither a directory or file at their root
            if not hasattr(module, "__file__"):
                continue
            # If the module only has one file, we only need to parse that file
            all_ast_nodes = parse_ast_for_file(
                python_file_path=getattr(module, "__file__")
            )
        else:
            # If the module has a directory root, we'll need to parse all .py files in the directory
            module_path = getattr(module, "__path__")
            if len(module_path) == 1:  # Most packaged modules have a source directory
                python_files = glob.glob(os.path.join(module_path[0], "**.py"))
                python_files += glob.glob(os.path.join(module_path[0], "**/*.py"))
            elif len(module_path) == 0:
                module_file = getattr(module, "__file__")
                python_files = [module_file]
            else:
                raise ValueError("Lug requires source code to get dependencies from module: ", module)

            for python_file_path in python_files:
                all_ast_nodes += parse_ast_for_file(python_file_path)

        # Now that we have the AST for all Python files in the module, we need to find the imports
        # There are two types of imports: direct imports and renamed imports. E.g.:
        # Direct import: import subprocess
        # Renamed import from subprocess import run
        ast_direct_imports = []
        ast_renamed_imports = []
        for node in all_ast_nodes:
            if isinstance(node, ast.Import):
                ast_direct_imports.append(node)
            if isinstance(node, ast.ImportFrom):
                ast_renamed_imports.append(node)

        # Both types of imports have different structures.
        # They're parsed differently, but added to the same list of imports once we find the module name.
        for renamed_import in ast_renamed_imports:
            module_name = renamed_import.module
            if module_name in filtered_sys_modules and module_name not in modules_to_skip:
                if "_pytest" in module.__name__:
                    continue
                deep_modules.add(sys.modules[module_name])
        for direct_import in ast_direct_imports:
            module_name = direct_import.names[0].name
            if module_name in filtered_sys_modules and module_name not in modules_to_skip:
                if "_pytest" in module.__name__:
                    continue
                deep_modules.add(sys.modules[module_name])
    return deep_modules


def search_for_modules_to_register(member, modules_to_register, discovered_module_names, depth=0):
    """Recursively search for modules to register."""
    if depth > 3:
        return
    is_module = isinstance(member, types.ModuleType)
    is_function = isinstance(member, types.FunctionType)
    if is_function:
        for global_var in member.__globals__.values():
            search_for_modules_to_register(
                member=global_var,
                modules_to_register=modules_to_register,
                discovered_module_names=discovered_module_names,
                depth=depth + 1
            )
        function_parent_module = inspect.getmodule(member)
        search_for_modules_to_register(
            member=function_parent_module,
            modules_to_register=modules_to_register,
            discovered_module_names=discovered_module_names,
            depth=depth + 1
        )
    elif is_module:
        module_name = getattr(member, "__name__")
        if module_name in discovered_module_names:
            return
        else:
            discovered_module_names.add(module_name)
        is_builtin_module = module_name in sys.builtin_module_names
        is_stdlib_module = module_name in STDLIB_MODULE_NAMES
        if not is_builtin_module and not is_stdlib_module:
            modules_to_register.append(member)
        if hasattr(member, "__globals__"):
            for global_var in member.__globals__.values():
                search_for_modules_to_register(
                    member=global_var,
                    modules_to_register=modules_to_register,
                    discovered_module_names=discovered_module_names,
                    depth=depth + 1,
                )


def get_modules_to_register(
        func,
        deep=False,
        modules_to_skip=frozenset({"pytest", "lug", "cloudpickle", "docker", "toolchest_client", "importlib_metadata"})
):
    modules_to_register = []
    search_for_modules_to_register(
        member=func,
        modules_to_register=modules_to_register,
        discovered_module_names=set(modules_to_skip),  # Skips unpicklable internal modules
    )
    deduplicated_modules_to_register = set(modules_to_register)
    if not deep:
        return deduplicated_modules_to_register
    additional_modules = get_registered_module_imports(
        deduplicated_modules_to_register,
        modules_to_skip=modules_to_skip,
    )
    return deduplicated_modules_to_register.union(additional_modules)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['lug']

package_data = \
{'': ['*']}

install_requires = \
['cloudpickle>=2.2.0,<3.0.0',
 'docker>=6.0.0,<7.0.0',
 'toolchest-client>=0.11.9,<0.12.0']

setup_kwargs = {
    'name': 'lug',
    'version': '0.2.7',
    'description': 'Run Python functions paired with any Docker container. Run locally or in the cloud.',
    'long_description': '# Lug\n\n<p align="center">\n    <a href="https://pypi.python.org/pypi/lug/" alt="PyPI version">\n        <img alt="PyPI" src="https://img.shields.io/pypi/v/lug?labelColor=212121&color=304FFE"></a>\n    <a href="https://github.com/trytoolchest/lug/" alt="Build">\n        <img src="https://img.shields.io/circleci/build/gh/trytoolchest/lug/main?label=build&token=3eb013dde86ed79996a768ab325cd30ea3a1c993&labelColor=212121&color=304FFE" /></a>\n    <a href="https://discord.gg/zgeJ9Pss" alt="Discord">\n        <img src="https://img.shields.io/discord/1016544715128176721?labelColor=212121&color=304FFE&label=discord" /></a>\n</p>\n\n**Lug** is an open source package that redirects Python calls to `subprocess.run`, `subprocess.Popen`, and `os.system` into \nany Docker container. This makes these system-level Python calls behave the same way on different machines, without \nrequiring any changes to the Docker container.\n\nLug also packages the Python function and Docker container, so you can easily run both in the cloud via Lug if needed. \nThis lets you give more computing power to the functions that need it.\n## Highlights\n\n- 📦 `subprocess.run`, `subprocess.Popen`, and `os.system` run in your container – for the dependencies you can\'t \ninstall with pip\n- 🐍 Use containers as-is, no need to add your Python version\n- ☁️ Run on your computer or in the cloud\n\n## Why use Lug?\n\nSome software – especially in bioinformatics and machine learning – can\'t be installed with `pip install`. Conda works \nin some cases, but for others a Docker container is the best solution. Lug makes the Docker container easier to work \nwith from Python.\n\nThe same software is usually resource intensive. Lug can run on your computer, or it can automatically run the Python \nfunction and Docker container in the cloud with [Toolchest](https://docs.trytoolchest.com/docs/pricing).\n\n## Prerequisites\n\n- macOS or Linux (supporting `POSIX_SPAWN`)\n- [Docker Engine](https://docs.docker.com/engine/install/) and the Docker CLI\n\n## Install\n\n[Install Docker Engine and the CLI](https://docs.docker.com/engine/install/), if you don\'t have it.\n\n### With pip:\n\n`pip install lug`\n\n### With Poetry:\n\n`poetry add lug`\n\n## Get started \n\n### Run a Python function locally in a Docker image\n\nEverything but the `subprocess.run` below runs as-is, with the `echo` command running in the Docker image:\n\n```python\nimport lug\nimport subprocess\n\n@lug.run(image="alpine:3.16.2")\ndef hello_world():\n    result = subprocess.run(\'echo "Hello, `uname`!"\', capture_output=True, text=True, shell=True)\n    return result.stdout\n\nprint(hello_world())\n```\n\nThat\'s it! After it finishes, you\'ll see `Hello, Linux!`.\n\n\n## Docs\n\nFull docs are at [lug.dev](https://lug.dev)\n\n## Open-source roadmap\n\n- [x] Run a Python function in a local container\n- [x] Maintain Python major.version in function and in container\n- [x] Serialize and deserialize Python function and Python dependencies\n- [x] `os.system()`, `subprocess.run()`, and `subprocess.Popen()` redirect to a user-specified container\n- [x] Local files passed to as input go to `./input/` in remote Docker container\n- [x] Remote files written to `./output/` in the container are written to local output Path\n- [x] Runs locally\n- [x] Run in the cloud with [Toolchest](https://github.com/trytoolchest/toolchest-client-python)\n- [ ] Stream live `stdout` during remote execution\n- [ ] `pip`-based environment propagation (help needed)\n- [ ] `conda`-based environment propagation (help needed)\n- [ ] Run in the cloud with AWS (help needed)\n- [ ] Run in the cloud with GCP (help needed)\n\n## License\n\nLug is licensed under the Apache License 2.0',
    'author': 'Justin Herr',
    'author_email': 'justin@trytoolchest.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://lug.dev',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)
